#from Pyqt5 user interface packages
from auxiliary.front_end_ui import Ui_MainWindow
from auxiliary.mask_selection import Ui_Dialog
from PyQt5.QtWidgets import QFileDialog
from qtpy import QtWidgets
from pyqtgraph.Qt import QtGui
import qdarkstyle

#custom made integration functions
from plot import plot_XZ_XY
from sim import VP, no_mask, custom

#usual packages
import numpy as np
import os
from matplotlib import pyplot as plt
import time
import sys
import configparser
import config

'''
User interface class that allows seting up parameters, running simulations and saving the obtained fields as .txt files
'''

class UI(QtGui.QMainWindow,Ui_MainWindow):
    def __init__(self, *args, **kwargs):
        print('PyFocus is running')    
        time.sleep(0.2)#so that tqdm can write
        QtWidgets.QMainWindow.__init__(self, *args, **kwargs)
        app = QtGui.QApplication.instance()
        app.setStyleSheet(qdarkstyle.load_stylesheet())#set a dark style

        #Main functions:
        self.counter=0# counts the numer of simulations realized in this sesion, to avoid overwriting save files with the same name
        self.setupUi(self)
        self.pushButton.clicked.connect(self.save_parameters)
        self.pushButton_2.clicked.connect(self.save_intensity)
        self.pushButton_3.clicked.connect(self.simulate)
        self.pushButton_4.clicked.connect(self.save_amplitudes)        
        self.pushButton_5.clicked.connect(self.selectsavefolder)   
        self.pushButton_6.clicked.connect(self.clear_plots)   
        self.radioButton.toggled.connect(self.lineEdit_19.setEnabled)#enable modifying parameters for simulation of the field inciding on the lens
        self.radioButton.toggled.connect(self.lineEdit_22.setEnabled)
        self.radioButton_2.toggled.connect(self.lineEdit_23.setEnabled)#enable modifying parameters for an interface
        self.radioButton_2.toggled.connect(self.lineEdit_24.setEnabled)
        self.radioButton_2.toggled.connect(self.lineEdit_25.setEnabled)
        self.radioButton_2.toggled.connect(self.lineEdit_2.setDisabled)
        
        #Functions dedicated to the saved files's names
        self.lineEdit_21.textEdited.connect(self.change_saving_name)
        self.comboBox.currentTextChanged.connect(self.change_default_name_and_open_dialog)
        self.modified_saving_name=False #Used later to avoid overwriting the changes in the name of the figure made by the user
        self.default_file_name='VP mask simulation' #default name for simulations if nothing is changed
        self.previous_figure_name=''#used to aoid overwriting previous figures which have the same name
        
    def selectsavefolder(self):
        '''
        Allows selecting a save folder for the txt files
        '''
        self.save_folder=QFileDialog.getExistingDirectory(self, "Select Directory")+'/'
        self.lineEdit_14.setText(self.save_folder)

    def get_parameters(self):
        '''
        Read parameters from the UI
        '''
        NA=float(self.lineEdit.text()) #numerical aperture
        n=float(self.lineEdit_2.text())#medium's density
        h=float(self.lineEdit_18.text()) #radius of aperture converted to nm
        w0=float(self.lineEdit_16.text()) #incident gaussian beam's radius, converted to nm
        wavelength=float(self.lineEdit_3.text())#wavelength in vacuum
        I0=float(self.lineEdit_15.text())#maximum intensity for the incident beam (|A|^2)
        zp0=float(self.lineEdit_12.text())#axial distance in wich to calculate the XY plane (z=cte=zp0 is the plane)
        gamma=float(self.lineEdit_11.text()) #arctan(ey/ex)
        beta=float(self.lineEdit_13.text()) #Delta phase betheen ey and ey, gamma=45 and beta=90 give right circular poalarizacion and a donut, gamma=45 and beta=-90 give left polarization and a maximum of intensity in the focused field
        x_steps=float(self.lineEdit_5.text())#radial pixel size
        z_steps=float(self.lineEdit_4.text())#axial pixel size
        x_range=float(self.lineEdit_17.text())#radial distance for the simulation
        if self.radioButton.isChecked()==True:#L and R are parameters for calculating the inciding beam with fraunhoffer's formula, if the propagation is depreciated then this parameters ar saved as ''not implemented''
            L=float(self.lineEdit_19.text())
            R=float(self.lineEdit_22.text())
        else:
            L='Not used'
            R='Not used'
        z_range=float(self.lineEdit_20.text())#axial distance for the simulation
        figure_name=str(self.lineEdit_21.text())#name for the ploted figures and the save files
        if figure_name=='': #to always have a figure name
            figure_name='Simulation'
        if self.radioButton_2.isChecked()==True:#if the interface simulation is not checked, this parameters ar saved as ''not implemented''
            try:
                z_int=float(self.lineEdit_25.text())#axial distance of the interface (negatie means before the focal plane)
                n=np.array([complex(a) for a in self.lineEdit_23.text().split(',')])#array containing the refraction indexes of all mediums
                ds=[]#array cointaining the thickness of all mediums
                ds.append(np.inf)
                if not self.lineEdit_24.text()=='':
                    for d in [float(s) for s in self.lineEdit_24.text().split(',')]:
                        ds.append(d)
                ds.append(np.inf)
                ds=np.array(ds)
            except:
                print('wrong format, explample of correct format: ns=1.5,1.4,1.4+1.4j,1.2j,1.33, ds=5,10,2',sys.exc_info())
            if not len(ds)==len(n):
                print('The refraction index array or the thickness array are missing a parameter, check that there are 2 parameters less for thickness than for refraction index')
                raise ValueError('The refraction index array or the thickness array are missing a parameter, check that there are 2 parameters less for thickness than for refraction index')
        else:
            z_int='Not used'
            ds='Not used'
        self.parameters=np.array((NA,n,h,w0,wavelength,gamma,beta,zp0,x_steps,z_steps,x_range,z_range,I0,L,R,ds,z_int,figure_name), dtype=object)

    def save_parameters(self):
        '''
        Save a txt file with the parameters used for the simulation with the UI, parameters can be strings
        '''
        try:
            folder=str(self.lineEdit_14.text())        
            if str(folder)=='(leave like this for local folder)':
                folder=''
            name=self.figure_name
            if name.endswith('.txt')==False: #to ensure that if the user writes .txt on the name, it doesnt get repeated as .txt.txt
                name=name+' parameters'
            else:
                name=name[:-4]+' parameters'
                        
            config = configparser.ConfigParser()
            # #check if an interface as simualted:
            # if len(self.parameters[1])==1:#no interface

            config['Parameters'] = {
            'NA': self.parameters[0],
            'Aperture radius (mm)':self.parameters[2],
            # 'focal distance (mm)': self.parameters[3],#removed since it is given by sine's law
            'incident beam radius (mm)': self.parameters[3],
            'wavelength at vacuum (nm)': self.parameters[4],
            'gamma, arctan(ey ex) (degrees)': self.parameters[5],
            'beta, Delta phase (degrees)': self.parameters[6],
            'Axial distance from focus for the XY plane (nm)': self.parameters[7],
            'Radial pixel size (nm)': self.parameters[8],
            'Axial pixel size (nm)': self.parameters[9],
            'Radial field of view (nm)': self.parameters[10],
            'Axial field of view (nm)': self.parameters[11],
            'Laser intensity (mW/cm^2)': self.parameters[12],
            'refraction indices': self.parameters[1],
            'interface layer thickness (nm)': self.parameters[15],
            'interface axial position (nm)': self.parameters[16],
            'Distance from phase plate to objective (mm)': self.parameters[13],
            'Phase mask radius (mm)': self.parameters[14]}

            with open(folder+name+'.txt', 'w') as configfile:
                config.write(configfile)
            print('Parameters saved!')
                 
        except:
            print("Unexpected error:", sys.exc_info())
            
    def save_intensity(self):
        '''
        Save a txt file with the intensity on the simulated planes
        '''
        try:
            folder=str(self.lineEdit_14.text())      
            if str(folder)=='(leave like this for local folder)':
                folder=''
            name=self.figure_name
            if name.endswith('.txt')==False: #to ensure that if the user writes .txt on the name, it doesnt get repeated as .txt.txt
                name_XZ=name+' XZ plane intensity.txt'
                name_XY=name+' XY plane intensity.txt'
            else:
                name_XZ=name[:-4]+' XZ plane intensity.txt'
                name_XY=name[:-4]+' XY plane intensity.txt'
            np.savetxt(folder+name_XY,self.Ifield_xy, fmt='%.18g', delimiter='\t', newline=os.linesep)        
            np.savetxt(folder+name_XZ,self.Ifield_xz, fmt='%.18g', delimiter='\t', newline=os.linesep)        
            print('Intensity saved!')
        except:
            print("Unexpected error:", sys.exc_info())

    def save_amplitudes(self):
        '''
        Save a txt file with the amplitude of each cartesian component on the simulated planes
        '''
        try:
            folder=str(self.lineEdit_14.text())      
            if str(folder)=='(leave like this for local folder)':
                folder=''
            name=self.figure_name
            if name.endswith('.txt')==False: #to ensure that if the user writes .txt on the name, it doesnt get repeated as .txt.txt
                name_x=name+' X component amplitude.txt'
                name_y=name+' Y component amplitude.txt'
                name_z=name+' Z component amplitude.txt'
            else:
                name_x=name[:-4]+' X component amplitude.txt'
                name_y=name[:-4]+' Y component amplitude.txt'
                name_z=name[:-4]+' Z component amplitude.txt'
            np.savetxt(folder+name_x,self.amplitudes_xy[0], fmt='%.18g', delimiter='\t', newline=os.linesep)        
            np.savetxt(folder+name_y,self.amplitudes_xy[1], fmt='%.18g', delimiter='\t', newline=os.linesep)        
            np.savetxt(folder+name_z,self.amplitudes_xy[2], fmt='%.18g', delimiter='\t', newline=os.linesep)        
            print('Amplitudes saved!')
        except:
            print("Unexpected error:", sys.exc_info())
      
    def clear_plots(self):
        '''
        Clear all existing figures
        '''
        plt.close('all')
        
    def change_saving_name(self):
        '''
        Change saving name for the UI, also modifying the 
        '''
        self.modified_saving_name=True
        self.name=self.lineEdit_21.text()
        
    def change_default_name_and_open_dialog(self):
        '''
        Quality of life modification, if a new mask is selected and the name has never been modified, then change the default simulation name 
        '''
        selected=self.comboBox.currentIndex()
        if self.modified_saving_name==False:
            if selected==0:#VP mask
                self.default_file_name='VP mask simulation'
                self.lineEdit_21.setText(self.default_file_name)
            if selected==1:#No mask (gaussian beam)
                self.default_file_name='Gaussian beam simulation'
                self.lineEdit_21.setText(self.default_file_name)
            if selected==2:#Custom mask
                self.default_file_name='Custom mask simulation'
                self.lineEdit_21.setText(self.default_file_name)
                # app = QtWidgets.QApplication(sys.argv)
                Dialog = QtWidgets.QDialog()
                ui2 = Ui_Dialog()
                ui2.setupUi(Dialog)
                Dialog.exec_()
    
    def simulate(self):
        '''
        Simulate with the UI using the functions provided in "sim"
        '''
        self.counter+=1
        self.get_parameters()
        x_range=self.parameters[10]
        z_range=self.parameters[11]
        figure_name=self.parameters[-1]
        #used to avoid overwriting previous figures which have the same name:
        if figure_name==self.default_file_name:
            figure_name+=' '+str(self.counter)
        if figure_name==self.previous_figure_name:
            figure_name+=' '+str(self.counter)
        self.previous_figure_name=figure_name        
        self.figure_name=figure_name #used as name for saving texts 
        
        selected=self.comboBox.currentIndex()
        
        try:      
            propagation=self.radioButton.isChecked() #then no need to calculate the field at the entrance of the lens
            interface=self.radioButton_2.isChecked()
            if selected==0: #VP mask
                #calculate field at the focal plane:
                ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY=VP(propagation,interface,*self.parameters)                
                    
            if selected==1: #No mask (gaussian beam)
                #calculate field at the focal plane:
                ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY=no_mask(propagation,interface,*self.parameters)
        
            if selected==2: #Custom mask
                if config.y==True: #internal variable used to check if given mask function is a functionor a matrix
                    aux='self.mask_function=lambda rho,phi,w0,f,k:'+config.x                       
                    exec(aux)
                else:
                    self.mask_function=config.x
                #calculate field at the focal plane:
                ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY=custom(self.mask_function,propagation,interface,*self.parameters)

            #plot the fields at the focus:
            plot_XZ_XY(ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY,x_range,z_range,figure_name) #ex, ey and ez have tha values of the field on the XY plane
            
            #make internal variables to save the fields as txts
            self.amplitudes_xy=(ex_XY,ey_XY,ez_XY)
            self.Ifield_xy=np.abs(ex_XY)**2+np.abs(ey_XY)**2+np.abs(ez_XY)**2#used to generate the save file
            self.Ifield_xz=np.zeros((2,2))#since it is not calculated i return a matrix of zeros so there is no error while trying to save the data
        except:
            print("Unexpected error:", sys.exc_info())

    def closeEvent(self, event, *args, **kwargs):
        '''
	Close the user interface
	'''
        print('PyFocus is closed')        
        super().closeEvent(*args, **kwargs)

        
        
if __name__ == '__main__':        
    gui=UI()
    gui.show()



import numpy as np
from matplotlib import pyplot as plt
from auxiliary.plot_polarization_elipses import polarization_elipse

def plot_XZ_XY(ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY,x_range,z_range,figure_name=''):
    '''
    Plot the calculated fields ont the XY and XZ planes
    
    Args:    
        :ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY (arrays): arrays with the amplitude of each cartesian component on the XZ plane (ex_XZ,ey_XZ,ez_XZ) or on the XY plane (ex_XY,ey_XY,ez_XY)
    
    Each index of the matrixes corresponds to a different pair of coordinates, for example:
        
    ex_XZ[z,x] with z each index of the coordinates np.linspace(z_range/2,-z_range/2,2*int(z_range/zsteps/2)) and x each index for np.linspace(-x_range/2**0.5,x_range/2**0.5,2*int(x_range/rsteps/2**0.5)) in which the field is calculated
    
    ex_XZ[y,x2] with y each index of the coordinates np.linspace(x_range/2,-x_range/2,2*int(x_range/rsteps/2)) and x each index for np.linspace(-x_range/2,x_range/2,2*int(x_range/rsteps/2)) in which the field is calculated
    
    The intensity is ploted in (kW/cm^2) since most focused fields at NA=1.4 have a maximum intensity in the order of 10^6 to 10^8 mW/cm^2
    
        :x_range,z_range: Range in the x and z coordinates respectively in which to plot the XZ and XY planes
    
    Returns;
        :fig1 (matplotlib figure): Figure showing the intensity on the XZ plane, the XY plane, the X axis and the polarization on the XY plane
        
        :fig2 (matplotlib figure): Figure showing the intensity and phase of each cartesian component on the XY plane
    '''
    #For pasage from (mW/cm^2) to (kW/cm^2) the intensity will be divided by 10**6
    
    zmax=z_range/2
    rmax=x_range*2**0.5/2 #the maximum radial distance is calculated sqrt(2) times biger than the maximum x or y distnace in the previous functions
    x,y=np.shape(ex_XZ)

    radial_pixel_width=x_range*2**0.5/2/y#value used to show the pixels centered at the radial position at which they are calculated
    Ifield_xz=np.abs(ex_XZ)**2+np.abs(ey_XZ)**2+np.abs(ez_XZ)**2
    Ifield_xz/=10**6
    
    plt.rcParams['font.size']=14
    #intensity plot
    fig1 = plt.figure(num=str(figure_name)+'_Intensity',figsize=(16, 8))
    spec = fig1.add_gridspec(ncols=3, nrows=2)
    ax1 = fig1.add_subplot(spec[0, 0])
    ax1.set_title('Normalized intensity',pad=20)
    pos=ax1.imshow(Ifield_xz,extent=[-rmax-radial_pixel_width,rmax-radial_pixel_width,-zmax,zmax], interpolation='none', aspect='equal')
    ax1.set_xlabel('x (nm)')
    ax1.set_ylabel('z (nm)')
    cbar1= fig1.colorbar(pos, ax=ax1)
    cbar1.ax.set_ylabel('Intensity (kW/cm\u00b2)')

    
    x2,y2=np.shape(ex_XY)
    Ifield_xy=np.abs(ex_XY)**2+np.abs(ey_XY)**2+np.abs(ez_XY)**2
    Ifield_xy/=10**6

    xmax=x_range/2
    extent=[-xmax-radial_pixel_width,xmax-radial_pixel_width,-xmax+radial_pixel_width,xmax+radial_pixel_width]
    ax2 = fig1.add_subplot(spec[0, 1])
    ax2.set_title('Intensity on xy',pad=20)
    pos2=ax2.imshow(Ifield_xy,extent=extent,interpolation='none', aspect='auto')
    cbar2=fig1.colorbar(pos2, ax=ax2)
    ax2.set_xlabel('x (nm)')
    ax2.set_ylabel('y (nm)')  
    ax2.axis('square')
    cbar2.ax.set_ylabel('Intensity (kW/cm\u00b2)')
    
    x2=np.shape(Ifield_xy)[0]
    Ifield_axis=Ifield_xy[int(x2/2),:]
    axis=np.linspace(-xmax-radial_pixel_width,xmax-radial_pixel_width,x2)
    ax3 = fig1.add_subplot(spec[0, 2])
    ax3.set_title('Intensity along x',pad=20)
    ax3.plot(axis,Ifield_axis)
    ax3.grid(True)
    ax3.set_xlabel('x (nm)')
    ax3.set_ylabel('Intensity  (kW/cm\u00b2)')  
    
    ax4 = fig1.add_subplot(spec[1, 1])
    ax4.set_title('Polarization on xy')
    pos4=ax4.imshow(Ifield_xy,extent=extent,interpolation='none', aspect='auto',alpha=0.5)
    cbar4=fig1.colorbar(pos4, ax=ax4)
    cbar4.ax.set_ylabel('Intensity (kW/cm\u00b2)')
    ax4.set_xlabel('x (nm)')
    ax4.set_ylabel('y (nm)')  
    ax4.axis('square')
    x_pos=np.linspace(-xmax*0.95,xmax*0.95,10)
    y_pos=np.linspace(-xmax*0.95,xmax*0.95,10)
    x_values=np.linspace(-xmax,xmax,np.shape(ex_XY)[0])
    y_values=np.linspace(xmax,-xmax,np.shape(ex_XY)[0])
    AMP=np.abs(xmax/6)
    for x_coor in x_pos:
        for y_coor in y_pos:
            x_index = (np.abs(x_values - x_coor)).argmin()
            y_index = (np.abs(y_values - y_coor)).argmin()
            # if np.abs(Ifield_xy[y_index,x_index]/np.max(Ifield_xy))>0.15 and np.abs(ez_XY[y_index,x_index])**2/np.max(Ifield_xy)<0.9:#added a condition for the z intensity to avoid point where the polarization is mostly along z
            if np.abs(Ifield_xy[y_index,x_index]/np.max(Ifield_xy))>0.15: #removed the z intensity condition
                if x_coor==x_pos[5] and y_coor==y_pos[5]: #at x,y=0,0 usually the atan2 does not work well
                    x_index+=1
                    y_index+=1
                polarization_elipse(ax4,x_coor,y_coor,ex_XY[y_index,x_index],ey_XY[y_index,x_index],AMP)
    '''
    #formula to calculate fwhm
    fwhm=np.abs(2*x_values[np.where(np.abs(Ifield_axis-np.max(Ifield_axis)/2)<0.05*np.max(Ifield_xy))[0]])
    print(fwhm)
    '''
    fig1.tight_layout()
    fig1.subplots_adjust(top=0.90)

    #Amplitud de  and fase plot 
    Amp_max=np.abs(np.max([np.max(np.abs(ex_XY)),np.max(np.abs(ey_XY)),np.max(np.abs(ez_XY))]))**2
    plt.rcParams['font.size']=14
    #ex
    fig2, ((ax_x1,ax_y1,ax_z1),(ax_x2,ax_y2,ax_z2)) = plt.subplots(num=str(figure_name)+'_Amplitude',figsize=(15, 8),nrows=2, ncols=3)
    ax_x1.set_title('$|E_{f_x}|^2$')
    pos_x1=ax_x1.imshow(np.abs(ex_XY)**2/Amp_max,extent=extent, interpolation='none', aspect='auto')
    ax_x1.set_xlabel('x (nm)')
    ax_x1.set_ylabel('y (nm)')
    ax_x1.axis('square')
    cbar_1_1=fig2.colorbar(pos_x1, ax=ax_x1)
    cbar_1_1.ax.set_ylabel('Relative intensity')
    
    ax_x2.set_title('$E_{f_x}$ phase')
    pos_x2=ax_x2.imshow(np.angle(ex_XY, deg=True)+180,extent=extent, interpolation='none', aspect='auto')
    ax_x2.set_xlabel('x (nm)')
    ax_x2.set_ylabel('y (nm)')
    ax_x2.axis('square')
    cbar_1_1=fig2.colorbar(pos_x2, ax=ax_x2,ticks=[5, 90,180,270,355])
    cbar_1_1.ax.set_ylabel('Phase (degrees)')

    #ey
    ax_y1.set_title('$|E_{f_y}|^2$')
    pos_y1=ax_y1.imshow(np.abs(ey_XY)**2/Amp_max,extent=extent, interpolation='none', aspect='auto')
    ax_y1.set_xlabel('x (nm)')
    ax_y1.set_ylabel('y (nm)')
    ax_y1.axis('square')
    cbar_1_1=fig2.colorbar(pos_y1, ax=ax_y1)
    cbar_1_1.ax.set_ylabel('Relative intensity')
    
    ax_y2.set_title('$E_{f_y}$ phase')
    pos_y2=ax_y2.imshow(np.angle(ey_XY, deg=True)+180,extent=extent, interpolation='none', aspect='auto')
    ax_y2.set_xlabel('x (nm)')
    ax_y2.set_ylabel('y (nm)')
    ax_y2.axis('square')
    cbar_1_1=fig2.colorbar(pos_y2, ax=ax_y2,ticks=[5, 90,180,270,355])
    cbar_1_1.ax.set_ylabel('Phase (degrees)')

    #ez
    ax_z1.set_title('$|E_{f_z}|^2$')
    pos_z1=ax_z1.imshow(np.abs(ez_XY)**2/Amp_max,extent=extent, interpolation='none', aspect='auto')
    ax_z1.set_xlabel('x (nm)')
    ax_z1.set_ylabel('y (nm)')
    ax_z1.axis('square')
    cbar_1_1=fig2.colorbar(pos_z1, ax=ax_z1)
    cbar_1_1.ax.set_ylabel('Relative intensity')
    
    ax_z2.set_title('$E_{f_z}$ phase')
    pos_z2=ax_z2.imshow(np.angle(ez_XY, deg=True)+180,extent=extent, interpolation='none', aspect='auto')
    ax_z2.set_xlabel('x (nm)')
    ax_z2.set_ylabel('y (nm)')
    ax_z2.axis('square')
    cbar_1_1=fig2.colorbar(pos_z2, ax=ax_z2,ticks=[5, 90,180,270,355])
    cbar_1_1.ax.set_ylabel('Phase (degrees)')
    fig2.tight_layout()
    
    return fig1,fig2




'''
High-level functions to simulate the foci obtained when: Focusing a gaussian beam, focusing a gaussian beam modulated with a VP mask and focusing a gaussian beam modulated with a custom phase mask

The functions use the 'parameters' array, defined as:
    
    parameters=np.array((NA, n, h, w0, wavelength, gamma, beta, z, x_steps, z_steps, x_range, z_range, I0, L, R, ds, z_int, figure_name), dtype=object)
    
With:
    
        :NA (float): Numerical aperture
        
        :n (float or array): Type depends on interface: float for no interface, array for interface. Refraction index for the medium of the optical system.
        
        :h (float): Radius of aperture of the objective lens(mm)
                
        :w0 (float): Radius of the incident gaussian beam (mm)
        
        :wavelength (float): Wavelength in the medium of the objective lens (equals to wavelength in vacuum/n)
        
        :gamma (float): Parameter that determines the polarization, arctan(ey/ex) (gamma=45, beta=90 gives right circular polarization and a donut shape)
        
        :beta (float): parameter that determines the polarization, phase difference between ex and ey (gamma=45, beta=90 gives right circular polarization and a donut shape)
        
        :z (float): Axial position for the XY plane (nm)
        
        :x_steps (float): Resolution in the X or Y coordinate for the focused field (nm)
        
        :z_steps (float): Resolution in the axial coordinate (Z) for the focused field (nm)
        
        :x_range (float): Field of view in the X or Y coordinate in which the focused field is calculated (nm)
        
        :z_range (float): field of view in the axial coordinate (z) in which the focused field is calculated (nm)
        
        :I_0 (float): Incident field intensity (mW/cm^2)
        
        :L (float, optional): Distance between phase mask and objective lens (mm), only used if propagation=True
        
        :R (float, optional): Phase mask radius (mm), only used if propagation=True
        
        :ds (array, optional): Thickness of each interface in the multilayer system (nm), must be given as a numpy array with the first and last values a np.inf. Only used if interface=True
        
        :z_int (float, optional): Axial position of the interphase. Only used if interface=True
        
        :figure_name (string, optional): Name for the images of the field. Also used as saving name if using the UI    
'''

import numpy as np
import time


from auxiliary.VP_functions import VP_integration, VP_fields, VP_fraunhofer
from auxiliary.no_mask_functions import no_mask_integration, no_mask_fields
from auxiliary.custom_mask_functions import generate_incident_field, plot_in_cartesian, custom_mask_objective_field, custom_mask_focus_field_XZ_XY
from auxiliary.interface import interface_custom_mask_focus_field_XZ_XY

def VP(propagation=False,interface=False,NA=1.4,n=1.5,h=3,w0=5,wavelength=640,gamma=45,beta=90,z=0,x_steps=5,z_steps=8,x_range=1000,z_range=2000,I0=1,L='',R='',ds='',z_int='',figure_name=''):
    '''
    Simulate the field obtained by focusing a gaussian beam modulated by a VP mask 
    
    Args:        
        :propagation (bool): True for calculating and ploting the field incident on the lens by fraunhofer's difractin formula, in which case R and L are needed; False for calculating the field incident on the lens while neglecting the propagation 
       
        :interface (bool): True for calculating the focused field with an interface, in which case ds and z_int are needed; Flase for calculating the field obtained without an interface
    
        :Parameters: NA,n,h,w0,wavelength,gamma,beta,z,x_steps,z_steps,x_range,z_range,I0,L,R,ds,z_int,figure_name: Simulation parameters, defined as part of the 'parameters' array
        
    Returns:        
        :arrays: ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY, each one is a matrix with the amplitude of each cartesian component on the XZ plane (ex_XZ,ey_XZ,ez_XZ) or on the XY plane (ex_XY,ey_XY,ez_XY)
    
        Each index of the matrixes corresponds to a different pair of coordinates, for example: 
            
        ex_XZ[z,x] with z each index of the coordinates np.linspace(z_range/2,-z_range/2,2*int(z_range/z_steps/2)) and x each index for np.linspace(-x_range/2**0.5,x_range/2**0.5,2*int(x_range/x_steps/2**0.5)) in which the field is calculated. 
        
        IMPORTANT: It is worth noting the x positions are sqrt(2) times higher to allow a square field of view for the XY plane (because the maximum radial position to be calculated for the XY plane is higher than the maximum x or y position)
        
        ex_XZ[y,x2] with y each index of the coordinates np.linspace(x_range/2,-x_range/2,2*int(x_range/x_steps/2)) and x each index for np.linspace(-x_range/2,x_range/2,2*int(x_range/x_steps/2)) in which the field is calculated
        
        The XZ plane is given by y=0 and the XY plane by the z parameter
        
        The unit of the intensity is the same as the initial intensity (mW/cm^2)
    '''
    
    
    if interface==False:#no interface
        alpha=np.arcsin(NA / n)
        f=h*n/NA#sine's law
        if propagation==False:
            II1,II2,II3,II4,II5=VP_integration(alpha,n,f,w0,wavelength/n,x_steps,z_steps,x_range,z_range)
            ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY=VP_fields(II1,II2,II3,II4,II5,wavelength/n,I0,gamma,beta,x_steps,z_steps,x_range,z_range,0,n,f,z)                        
        else:
            E_rho,Ex,Ey=VP_fraunhofer(gamma,beta,1000,R,L,I0,wavelength/n,2*h,w0,2000,20,plot=True,figure_name=figure_name)
            '''
            To calculate the focused field, doing a 2D integration with the trapezoid method will be faster than the 1D integration, since the function converges really slowly with scipy.integrate.quad, requiring an hour at least. 1D integration method is used by the funcions "VP_integration_with_propagation" and "VP_fields_with_propagation" in VP.py, but unimplemented here
            '''
            #Usual parameters for integration precision:
            #resolution for field at objective
            divisions_theta=200
            divisions_phi=200

            ex_lens=np.zeros((divisions_phi,divisions_theta),dtype=complex)
            ey_lens=np.zeros((divisions_phi,divisions_theta),dtype=complex)
        
            theta_values=np.linspace(0,alpha,divisions_theta)  #divisions of theta in which the trapezoidal 2D integration is done
            rhop_values=np.sin(theta_values)*f              #given by the sine's law
            phip_values=np.linspace(0,2*np.pi,divisions_phi)   #divisions of phi in which the trapezoidal 2D integration is done
            for i,phip in enumerate(phip_values):
                for j,rhop in enumerate(rhop_values):
                    ex_lens[i,j]=E_rho(rhop)*np.exp(1j*phip)
                    ey_lens[i,j]=E_rho(rhop)*np.exp(1j*phip)
            ex_lens*=np.cos(gamma*np.pi/180)*I0**0.5
            ey_lens*=np.sin(gamma*np.pi/180)*np.exp(1j*beta*np.pi/180)*I0**0.5

            #resolution for field near the focus, smaller numbers give faster integration times
            resolution_focus=int(x_range/x_steps)
            resolution_z=int(z_range/z_steps)
            
            #calculate field at the focal plane:
            ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY=custom_mask_focus_field_XZ_XY(ex_lens,ey_lens,alpha,h,wavelength/n,z_range,resolution_z,z,resolution_focus,divisions_theta,divisions_phi,x_range)
    
    else:#interface is given
        alpha=np.arcsin(NA / n[0])
        f=h*n[0]/NA    
        mask_function_VP=lambda theta, phi,w0,f,wavelength: np.exp(1j*phi)# is the VP mask function
        #Usual parameters for integration precision:
        #resolution for field at objective
        divisions_theta=200
        divisions_phi=200
        
        #resolution for field near the focus
        resolution_focus=int(x_range/x_steps)
        resolution_z=int(z_range/z_steps)
        
        #smaller numbers give faster integration times
        
        if propagation==False:
            #calculate field inciding on the lens by multiplying the phase mask's function and the gaussian amplitude
            #(ex_lens,ey_lens) are 2 matrixes with the values of the incident amplitude for each value in phi,theta                   
            ex_lens,ey_lens=generate_incident_field(mask_function_VP,alpha,f,divisions_phi,divisions_theta,gamma,beta,w0,I0,wavelength/n[0])            
            #plot field at the entrance of the lens:
            #since ex_lens and ey_lens are given in theta and phi coordinates, the have to be transformed to cartesian in order to be ploted, hence the name of this function
        else:
            #calculate field inciding on the lens by fraunhofer's difraction formula
            #N_rho and N_phi are the number of divisions for fraunhoffer's integral by the trapecium's method
            N_rho=400
            N_phi=400                
            ex_lens,ey_lens,I_cartesian,Ex_cartesian,Ey_cartesian=custom_mask_objective_field(h,gamma,beta,divisions_theta,divisions_phi,N_rho,N_phi,alpha,f,mask_function_VP,R,L,I0,wavelength,w0,figure_name,plot=True)

        #calculate field at the focal plane:
        ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY=interface_custom_mask_focus_field_XZ_XY(n,ds,ex_lens,ey_lens,alpha,h,wavelength,z_int,z_range,resolution_z,z,resolution_focus,divisions_theta,divisions_phi,x_range)
    
    return ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY

def no_mask(propagation=False,interface=False,NA=1.4,n=1.5,h=3,w0=5,wavelength=640,gamma=45,beta=90,z=0,x_steps=5,z_steps=8,x_range=1000,z_range=2000,I0=1,L='',R='',ds='',z_int='',figure_name=''):#f (the focal distance) is given by the sine's law, but can be modified if desired
    '''
    Simulate the field obtained by focusing a gaussian beam without being modulated in phase
    
    Args:        
        :propagation (bool): Kept for homogeneity of the functions, True yields no diference from False, where the field incident on the lens is calculated by neglecting the propagation towards the objective lens
        
        :interface (bool): True for calculating the focused field with an interface, in which case ds and z_int are needed; Flase for calculating the field obtained without an interface
    
        :Parameters: NA,n,h,w0,wavelength,gamma,beta,z,x_steps,z_steps,x_range,z_range,I0,L,R,ds,z_int,figure_name: Simulation parameters, defined as part of the 'parameters' array
        
    Returns:        
        :arrays: ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY, each one is a matrix with the amplitude of each cartesian component on the XZ plane (ex_XZ,ey_XZ,ez_XZ) or on the XY plane (ex_XY,ey_XY,ez_XY)
    
        Each index of the matrixes corresponds to a different pair of coordinates, for example: 
            
        ex_XZ[z,x] with z each index of the coordinates np.linspace(z_range/2,-z_range/2,2*int(z_range/z_steps/2)) and x each index for np.linspace(-x_range/2**0.5,x_range/2**0.5,2*int(x_range/x_steps/2**0.5)) in which the field is calculated. 
        
        IMPORTANT: It is worth noting the x positions are sqrt(2) times higher to allow a square field of view for the XY plane (because the maximum radial position to be calculated for the XY plane is higher than the maximum x or y position)
        
        ex_XZ[y,x2] with y each index of the coordinates np.linspace(x_range/2,-x_range/2,2*int(x_range/x_steps/2)) and x each index for np.linspace(-x_range/2,x_range/2,2*int(x_range/x_steps/2)) in which the field is calculated
        
        The XZ plane is given by y=0 and the XY plane by the z parameter 
        
        The unit of the intensity is the same as the initial intensity (mW/cm^2)
    '''
    
    if propagation==True:
        print('Propagation for field incident on the lens is only done for VP and custom masks. Calculating focused field with propagation=False:')
        time.sleep(0.5)
    if interface==False:#no interface
        alpha=np.arcsin(NA / n)
        f=h*n/NA
        II1,II2,II3=no_mask_integration(alpha,n,f,w0,wavelength/n,x_range,z_range,z_steps,x_steps)
        ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY=no_mask_fields(II1,II2,II3,wavelength/n,I0,beta,gamma,z_steps,x_steps,x_range,z_range,0,n,f,z)

    else:#interface
        f=h*n[0]/NA    
        alpha=np.arcsin(NA / n[0])
        mask_function=lambda theta, phi,w0,f,wavelength: 1# is the nule mask function
        divisions_theta=200
        divisions_phi=200
        resolution_focus=int(x_range/x_steps)
        resolution_z=int(z_range/z_steps)

        ex_lens,ey_lens=generate_incident_field(mask_function,alpha,f,divisions_phi,divisions_theta,gamma,beta,w0,I0,wavelength/n[0])
        ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY=interface_custom_mask_focus_field_XZ_XY(n,ds,ex_lens,ey_lens,alpha,h,wavelength,z_int,z_range,resolution_z,z,resolution_focus,divisions_theta,divisions_phi,x_range,x0=0,y0=0,z0=0,plot_plane='X')

    return ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY

def custom(mask_function, propagation=False,interface=False,NA=1.4,n=1.5,h=3,w0=5,wavelength=640,gamma=45,beta=90,z=0,x_steps=5,z_steps=8,x_range=1000,z_range=2000,I0=1,L='',R='',ds='',z_int='',figure_name='',divisions_theta=200,divisions_phi=200,plot_Ei=True):#f (the focal distance) is given by the sine's law, but can be modified if desired
    '''
    Simulate the field obtained by focusing a gaussian beam modulated by a custom phase mask
    
    Args:        
        :mask_function (array or function): Parameter defining the custom phase mask to be simulated:
            If array: Complex amplitude of the custom phase mask for each value of the spherical coordinates theta and phi: custom_mask[phi_position,theta_position] for phi_position a value in np.linspace(0,2*np.pi,divisions_phi) and theta_position a value in np.linspace(0,alpha,divisions_theta) 
                
            If function: Analytical function that defines the phase mask, must be a function of the 5 internal variables: rho, phi, w0, f and k, with:
                
            rho: Radial coordinate from 0 to the aperture radius of the objective.
            
            phi: Azimutal coordinate from 0 to 2pi.
            
            w0: Radius of the incident gaussian beam.
            
            f: Focal distane of the objective lens
            
            k: Wavenumber in the objective lens medium IMPORTANT:(mm)
            
        The real part defines the amplitude of the incident field
        
        :propagation (bool): True for calculating and ploting the field incident on the lens by fraunhofer's difractin formula, in which case R and L are needed; False for calculating the field incident on the lens while neglecting the propagation
        
        :interface (bool): True for calculating the focused field with an interface, in which case ds and z_int are needed; Flase for calculating the field obtained without an interface
    
        :Parameters: NA,n,h,w0,wavelength,gamma,beta,z,x_steps,z_steps,x_range,z_range,I0,L,R,ds,z_int,figure_name: Simulation parameters, defined as part of the 'parameters' array
    Returns:        
        :arrays: ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY, each one is a matrix with the amplitude of each cartesian component on the XZ plane (ex_XZ,ey_XZ,ez_XZ) or on the XY plane (ex_XY,ey_XY,ez_XY)
    
        Each index of the matrixes corresponds to a different pair of coordinates, for example: 
            
        ex_XZ[z,x] with z each index of the coordinates np.linspace(z_range/2,-z_range/2,2*int(z_range/z_steps/2)) and x each index for np.linspace(-x_range/2**0.5,x_range/2**0.5,2*int(x_range/x_steps/2**0.5)) in which the field is calculated. 
        
        IMPORTANT: It is worth noting the x positions are sqrt(2) times higher to allow a square field of view for the XY plane (because the maximum radial position to be calculated for the XY plane is higher than the maximum x or y position)
        
        ex_XZ[y,x2] with y each index of the coordinates np.linspace(x_range/2,-x_range/2,2*int(x_range/x_steps/2)) and x each index for np.linspace(-x_range/2,x_range/2,2*int(x_range/x_steps/2)) in which the field is calculated
        
        The XZ plane is given by y=0 and the XY plane by the z parameter 
        
        The unit of the intensity is the same as the initial intensity (mW/cm^2)
    '''
    #moving to wavelength in the optical system's medium
    if interface==False:
        alpha=np.arcsin(NA / n)
        wavelength/=n
        f=h*n/NA    
    else:
        alpha=np.arcsin(NA / n[0])
        wavelength/=n[0]
        f=h*n[0]/NA    

    # mask_function=lambda theta, phi: np.exp(1j*phi) is the VP mask function
    #Usual parameters for integration precision:
    
    #resolution for field near the focus
    resolution_focus=int(x_range/x_steps)
    resolution_z=int(z_range/z_steps)
    
    #smaller numbers give faster integration times
    
    if callable(mask_function)==True:#if a function is given, turn it into the ex_lens and ey_lens variables
        #resolution for field at objective
        if propagation==False:
            #calculate field inciding on the lens by multiplying the phase mask's function and the gaussian amplitude
            ex_lens,ey_lens=generate_incident_field(mask_function,alpha,f,divisions_phi,divisions_theta,gamma,beta,w0,I0,wavelength)
            #plot field at the entrance of the lens:
            #since ex_lens and ey_lens are given in theta and phi coordinates, the have to be transformed to cartesian in order to be ploted, hence the name of this function
            if plot_Ei==True:
                print('Showing incident field:')
                plot_in_cartesian(ex_lens,ey_lens,h,alpha,f,figure_name)
        else:
            #calculate field inciding on the lens by fraunhofer's difraction formula
            #N_rho and N_phi are the number of divisions for fraunhoffer's integral by the trapecium's method
            N_rho=200
            N_phi=200                
            ex_lens,ey_lens,I_cartesian,Ex_cartesian,Ey_cartesian=custom_mask_objective_field(h,gamma,beta,divisions_theta,divisions_phi,N_rho,N_phi,alpha,f,mask_function,R,L,I0,wavelength,w0,figure_name,plot=True)
            if plot_Ei==True:
                print('Showing incident field:')
                plot_in_cartesian(ex_lens,ey_lens,h,alpha,f,figure_name)

    elif type(mask_function)==np.ndarray:#if an array is given
        if propagation==True:
            print('Giving mask function as an array not implemented for calculation of incient field propagation. Can be implemented by obtaining "ex_lens" and "ey_lens" arrays using the "generate_incident_field" function manualy replacing "mask_function" with the desired array')
            print('Simulation will continue with propagation=False')
        ex_lens,ey_lens=mask_function*np.cos(gamma*np.pi/180)*I0**0.5,mask_function*np.sin(gamma*np.pi/180)*np.exp(1j*beta*np.pi/180)*I0**0.5 #make the x and y component based on polarization
        if plot_Ei==True:
            print('Showing incident field:')
            plot_in_cartesian(ex_lens,ey_lens,h,alpha,f,figure_name)
        divisions_phi,divisions_theta=np.shape(mask_function)
    else:
        print('Wrong format for mask function, acceptable formats are functions or arrays')
        
    #calculate field at the focal plane:
    if interface==False:#no interface
        ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY=custom_mask_focus_field_XZ_XY(ex_lens,ey_lens,alpha,h,wavelength,z_range,resolution_z,z,resolution_focus,divisions_theta,divisions_phi,x_range)
    else:
        wavelength*=n[0] #to return to wavelength at vacuum
        ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY=interface_custom_mask_focus_field_XZ_XY(n,ds,ex_lens,ey_lens,alpha,h,wavelength,z_int,z_range,resolution_z,z,resolution_focus,divisions_theta,divisions_phi,x_range)

    return ex_XZ,ey_XZ,ez_XZ,ex_XY,ey_XY,ez_XY

